"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const debug_1 = __importDefault(require("debug"));
const internals_1 = require("./internals");
const debugQuery = debug_1.default("squid:query");
function escapeIdentifier(identifier) {
    if (identifier.charAt(0) === '"' && identifier.charAt(identifier.length - 1) === '"') {
        identifier = identifier.substr(1, identifier.length - 2);
    }
    if (identifier.includes('"')) {
        throw new Error(`Invalid identifier: ${identifier}`);
    }
    return `"${identifier}"`;
}
function objectEntries(object, filterUndefinedValues) {
    const keys = Object.keys(object);
    const entries = keys.map(key => [key, object[key]]);
    return filterUndefinedValues
        ? entries.filter(([, columnValue]) => typeof columnValue !== "undefined")
        : entries;
}
function serializeSqlTemplateExpression(expression, nextParamID) {
    const values = [];
    if (internals_1.isSpecialExpression(expression)) {
        return expression.buildFragment(nextParamID);
    }
    else {
        const text = pushSqlParameter(nextParamID, values, expression);
        return {
            text,
            values
        };
    }
}
/**
 * SQL template string tag. Returns a query object: `{ text: string, values: any[] }`.
 * Template expressions are automatically SQL-injection-proofed unless wrapped in `sql.raw()`.
 *
 * @example
 * const { rows } = await database.query(sql`SELECT name, email FROM users WHERE id=${userID}`)
 * @example
 * const { rows } = await database.query<{ name: string, email: string }>(sql`
 *   SELECT name, email FROM users WHERE id=${userID}
 * `)
 */
function sql(texts, ...values) {
    let parameterValues = [];
    let resultingSqlQuery = texts[0];
    for (let valueIndex = 0; valueIndex < values.length; valueIndex++) {
        const expression = values[valueIndex];
        const followingSqlText = texts[valueIndex + 1];
        const nextParamID = parameterValues.length + 1;
        const serialized = serializeSqlTemplateExpression(expression, nextParamID);
        resultingSqlQuery += serialized.text + (followingSqlText || "");
        parameterValues = [...parameterValues, ...serialized.values];
    }
    const query = {
        text: resultingSqlQuery,
        values: parameterValues
    };
    debugQuery(query);
    return query;
}
exports.sql = sql;
/**
 * Wrap an SQL template expression value in `sql.raw()` to inject the raw value into the query.
 * Attention: Can easily lead to SQL injections! Use with caution and only if necessary.
 */
function rawExpression(rawValue) {
    return {
        type: internals_1.$sqlExpressionValue,
        buildFragment() {
            return {
                text: rawValue,
                values: []
            };
        }
    };
}
exports.raw = rawExpression;
function safeExpression(value) {
    return {
        type: internals_1.$sqlExpressionValue,
        buildFragment(paramID) {
            return {
                text: `\$${paramID}`,
                values: [value]
            };
        }
    };
}
exports.safe = safeExpression;
sql.raw = rawExpression;
sql.safe = safeExpression;
function pushSqlParameter(nextParamID, parameterValues, value) {
    parameterValues.push(value);
    return `\$${nextParamID}`;
}
function buildSpreadAndFragment(columnValues, nextParamID) {
    let values = [];
    const andChain = columnValues
        .map(([columnName, columnValue]) => {
        const serialized = serializeSqlTemplateExpression(columnValue, nextParamID);
        values = [...values, ...serialized.values];
        nextParamID += serialized.values.length;
        return `${escapeIdentifier(columnName)} = ${serialized.text}`;
    })
        .join(" AND ");
    return {
        text: `(${andChain})`,
        values
    };
}
function buildSpreadInsertFragment(columns, records, nextParamID) {
    let values = [];
    const text = `(${columns.map(columnName => escapeIdentifier(columnName)).join(", ")})` +
        ` VALUES ` +
        `(${records
            .map(record => {
            return columns
                .map(columnName => {
                const columnValue = record[columnName];
                if (typeof columnValue === "undefined") {
                    throw new TypeError(`Missing value for column "${columnName}"`);
                }
                const serialized = serializeSqlTemplateExpression(columnValue, nextParamID);
                values = [...values, ...serialized.values];
                nextParamID += serialized.values.length;
                return serialized.text;
            })
                .join(", ");
        })
            .join("), (")})`;
    return {
        text,
        values
    };
}
function buildSpreadUpdateFragment(updateValues, nextParamID) {
    let values = [];
    const settersChain = updateValues
        .map(([columnName, columnValue]) => {
        const serialized = serializeSqlTemplateExpression(columnValue, nextParamID);
        values = [...values, ...serialized.values];
        nextParamID += serialized.values.length;
        return `${escapeIdentifier(columnName)} = ${serialized.text}`;
    })
        .join(", ");
    return {
        text: `${settersChain}`,
        values
    };
}
/**
 * Extract colum names from a records.
 * Make sure all records has the same number of columns.
 */
function extractColumsName(records) {
    const referenceRecord = records[0];
    if (records.some(record => Object.keys(record).length !== Object.keys(referenceRecord).length)) {
        throw new Error("Rows must be of the same length");
    }
    return Object.keys(referenceRecord).filter(columnName => typeof referenceRecord[columnName] !== "undefined");
}
/**
 * Convenience function to keep WHERE clauses concise. Takes an object:
 * Keys are supposed to be a column name, values the values that the record must have set.
 *
 * @example
 * const { rows } = await database.query(sql`SELECT * FROM users WHERE ${spreadAnd({ name: "John", email: "john@example.com" })}`)
 */
function spreadAnd(record) {
    const values = objectEntries(record, true);
    return {
        type: internals_1.$sqlExpressionValue,
        buildFragment(nextParamID) {
            return buildSpreadAndFragment(values, nextParamID);
        }
    };
}
exports.spreadAnd = spreadAnd;
/**
 * Convenience function to keep INSERT statements concise.
 * @example
 * await database.query(sql`INSERT INTO users ${spreadInsert({ name: "John", email: "john@example.com" })}`)
 */
function spreadInsert(...records) {
    const colums = extractColumsName(records);
    return {
        type: internals_1.$sqlExpressionValue,
        buildFragment(nextParamID) {
            return buildSpreadInsertFragment(colums, records, nextParamID);
        }
    };
}
exports.spreadInsert = spreadInsert;
/**
 * Convenience function to keep UPDATE statements concise. Takes an object:
 * @example
 * await database.query(sql`UPDATE users SET ${spreadUpdate({ name: "John", email: "john@example.com" })} WHERE id = 1`)
 */
function spreadUpdate(record) {
    const values = objectEntries(record, true);
    return {
        type: internals_1.$sqlExpressionValue,
        buildFragment(nextParamID) {
            return buildSpreadUpdateFragment(values, nextParamID);
        }
    };
}
exports.spreadUpdate = spreadUpdate;
